/****************************************************************************
** Copyright (c) 2016, Fougue SAS <https://www.fougue.pro>
** SPDX-License-Identifier: BSD-2-Clause
****************************************************************************/

#pragma once

#include "../base/occ_handle.h"
#include <SelectMgr_EntityOwner.hxx>

namespace Mayo {

using GraphicsOwnerPtr = OccHandle<SelectMgr_EntityOwner>;

} // namespace Mayo
