#pragma once

#include <algorithm>
#include <array>
#include <filesystem>
#include <functional>
#include <iostream>
#include <memory>
#include <optional>
#include <string>
#include <string_view>
#include <type_traits>
#include <unordered_map>
#include <unordered_set>
#include <utility>
#include <variant>
#include <vector>

#include <cassert>
#include <cstdint>

#include <fmt/format.h>
#include <gsl/span>
#include <kdbindings/signal.h>
#include <magic_enum/magic_enum.hpp>

//#include <Standard_Handle.hxx>
//#include <TDF_Label.hxx>

#include <QtCore/QObject>
#include <QtCore/QString>
#include <QtCore/QVariant>
#include <QtWidgets/QWidget>
